@@ -17,6 +17,7 @@ import qualified Data.Binary as Binary
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.Char as Char
+import Data.Constraint (Dict(Dict))
 import Data.Maybe (catMaybes, fromJust, fromMaybe, isJust, isNothing)
 import Data.Proxy (Proxy(Proxy))
 import Data.Ratio ((%), numerator, denominator)
@@ -36,36 +37,27 @@ import qualified Test.Tasty.QuickCheck as QC
 
 import qualified Money
 import qualified Money.Internal as MoneyI
-  (rationalFromDecimal, rationalToDecimal)
 
 --------------------------------------------------------------------------------
 
--- | Generates a valid 'MoneyI.rationalToDecimal' result. Returns the thousand
--- and decimal separators as welland decimal separators as well.
-genDecimal :: QC.Gen (T.Text, Maybe Char, Char)
+-- | A wrapper around a decimal representation and its 'Money.DecimalConf',
+-- mostly just to benefit from its 'QC.Arbitrary' instance.
+data Decimal = Decimal Money.DecimalConf T.Text
+  deriving (Show)
+
+instance QC.Arbitrary Decimal where
+  arbitrary = uncurry (flip Decimal) <$> genDecimal
+  shrink (Decimal ds t) = Decimal <$> QC.shrink ds
+                                  <*> fmap T.pack (QC.shrink (T.unpack t))
+
+-- | Generates a valid 'MoneyI.rationalToDecimal' result using the returned
+-- 'Money.DecimalConf'.
+genDecimal :: QC.Gen (T.Text, Money.DecimalConf)
 genDecimal = do
   aprox :: Money.Approximation <- QC.arbitrary
-  plus :: Bool <- QC.arbitrary
-  digs :: Word8 <- QC.arbitrary
+  ds :: Money.DecimalConf <- QC.arbitrary
   r :: Rational <- (%) <$> QC.arbitrary <*> QC.suchThat QC.arbitrary (/= 0)
-  (yts, ds) <- genDecimalSeps
-  let
-      dec = fromMaybe
-          (error "genDecimal failed because rationalToDecimal returned Nothing")
-          (MoneyI.rationalToDecimal aprox plus yts ds digs r)
-  pure (dec, yts, ds)
-
--- | Generates valid separators for decimal representations (see genDecimal).
-genDecimalSeps :: QC.Gen (Maybe Char, Char)
-genDecimalSeps = do
-  let msep = QC.suchThat QC.arbitrary (not . Char.isDigit)
-  ds :: Char <- msep
-  yts :: Maybe Char <- genMaybe (QC.suchThat msep (/= ds))
-  pure (yts, ds)
-
-
-genMaybe :: QC.Gen a -> QC.Gen (Maybe a)
-genMaybe m = QC.oneof [pure Nothing, fmap Just m]
+  pure (MoneyI.rationalToDecimal ds aprox r, ds)
 
 --------------------------------------------------------------------------------
 
@@ -79,10 +71,12 @@ tests :: Tasty.TestTree
 tests =
   Tasty.testGroup "root"
   [ testCurrencies
+  , testHalfEvenRounding
   , testCurrencyUnits
   , testExchange
   , testRationalToDecimal
   , testRationalFromDecimal
+  , testMkSeparators
   , testDiscreteFromDecimal
   , testRawSerializations
   ]
@@ -99,10 +93,8 @@ testCurrencies =
 testCurrencyUnits :: Tasty.TestTree
 testCurrencyUnits =
   Tasty.testGroup "Currency units"
-  [ testDiscrete (Proxy :: Proxy "BTC") (Proxy :: Proxy "BTC")
-  , testDiscrete (Proxy :: Proxy "BTC") (Proxy :: Proxy "satoshi")
+  [ testDiscrete (Proxy :: Proxy "BTC") (Proxy :: Proxy "satoshi")
   , testDiscrete (Proxy :: Proxy "BTC") (Proxy :: Proxy "bitcoin")
-  , testDiscrete (Proxy :: Proxy "USD") (Proxy :: Proxy "USD")
   , testDiscrete (Proxy :: Proxy "USD") (Proxy :: Proxy "cent")
   , testDiscrete (Proxy :: Proxy "USD") (Proxy :: Proxy "dollar")
   , testDiscrete (Proxy :: Proxy "VUV") (Proxy :: Proxy "vatu")
@@ -202,6 +194,29 @@ testRationalToDecimal =
          , "0"       --  6
          , "0"       --  7
          ]
+  , HU.testCase "Round: r5" $ do
+       render Money.Round r5 @?=
+         [ "0.50"   --  0
+         , "0.50"   --  1
+         , "+0.50"  --  2
+         , "+0.50"  --  3
+         , "0"      --  4
+         , "0"      --  5
+         , "0"      --  6
+         , "0"      --  7
+         ]
+  , HU.testCase "Round: negate r5" $ do
+       render Money.Round (negate r5) @?=
+         [ "-0.50"   --  0
+         , "-0.50"   --  1
+         , "-0.50"   --  2
+         , "-0.50"   --  3
+         , "0"       --  4
+         , "0"       --  5
+         , "0"       --  6
+         , "0"       --  7
+         ]
+
   , HU.testCase "Floor: r1" $ do
        render Money.Floor r1 @?=
          [ "1023004567.89"        --  0
@@ -290,6 +305,28 @@ testRationalToDecimal =
          , "-1"      --  6
          , "-1"      --  7
          ]
+  , HU.testCase "Floor: r5" $ do
+       render Money.Floor r5 @?=
+         [ "0.50"   --  0
+         , "0.50"   --  1
+         , "+0.50"   --  2
+         , "+0.50"   --  3
+         , "0"      --  4
+         , "0"      --  5
+         , "0"      --  6
+         , "0"      --  7
+         ]
+  , HU.testCase "Floor: negate r5" $ do
+       render Money.Floor (negate r5) @?=
+         [ "-0.50"   --  0
+         , "-0.50"   --  1
+         , "-0.50"   --  2
+         , "-0.50"   --  3
+         , "-1"      --  4
+         , "-1"      --  5
+         , "-1"      --  6
+         , "-1"      --  7
+         ]
 
   , HU.testCase "Ceiling: r1" $ do
        render Money.Ceiling r1 @?=
@@ -379,6 +416,28 @@ testRationalToDecimal =
          , "0"      --  6
          , "0"      --  7
          ]
+  , HU.testCase "Ceiling: r5" $ do
+       render Money.Ceiling r5 @?=
+         [ "0.50"   --  0
+         , "0.50"   --  1
+         , "+0.50"  --  2
+         , "+0.50"  --  3
+         , "1"      --  4
+         , "1"      --  5
+         , "+1"     --  6
+         , "+1"     --  7
+         ]
+  , HU.testCase "Ceiling: negate r5" $ do
+       render Money.Ceiling (negate r5) @?=
+         [ "-0.50"   --  0
+         , "-0.50"   --  1
+         , "-0.50"   --  2
+         , "-0.50"   --  3
+         , "0"      --  4
+         , "0"      --  5
+         , "0"      --  6
+         , "0"      --  7
+         ]
 
   , HU.testCase "Truncate: r1" $ do
       render Money.Truncate r1 @?= render Money.Floor r1
@@ -403,50 +462,114 @@ testRationalToDecimal =
 
   , HU.testCase "Truncate: negate r4" $ do
       render Money.Truncate (negate r4) @?= render Money.Ceiling (negate r4)
+
+  , HU.testCase "Truncate: r5" $ do
+      render Money.Truncate r5 @?= render Money.Floor r5
+
+  , HU.testCase "Truncate: negate r5" $ do
+      render Money.Truncate (negate r5) @?= render Money.Ceiling (negate r5)
+
+  , HU.testCase "HalfEven: r1" $ do
+      render Money.HalfEven r1 @?= render Money.Round r1
+
+  , HU.testCase "HalfEven: negate r1" $ do
+      render Money.HalfEven (negate r1) @?= render Money.Round (negate r1)
+
+  , HU.testCase "HalfEven: r2" $ do
+      render Money.HalfEven r2 @?= render Money.Round r2
+
+  , HU.testCase "HalfEven: negate r2" $ do
+      render Money.HalfEven (negate r2) @?= render Money.Round (negate r2)
+
+  , HU.testCase "HalfEven: r3" $ do
+      render Money.HalfEven r3 @?= render Money.Round r3
+
+  , HU.testCase "HalfEven: negate r3" $ do
+      render Money.HalfEven (negate r3) @?= render Money.Round (negate r3)
+
+  , HU.testCase "HalfEven: r4" $ do
+      render Money.HalfEven r4 @?= render Money.Round r4
+
+  , HU.testCase "HalfEven: negate r4" $ do
+      render Money.HalfEven (negate r4) @?= render Money.Round (negate r4)
+
+  , HU.testCase "HalfEven: r5" $ do
+       render Money.HalfEven r5 @?=
+         [ "0.50"   --  0
+         , "0.50"   --  1
+         , "+0.50"  --  2
+         , "+0.50"  --  3
+         , "0"      --  4
+         , "0"      --  5
+         , "0"      --  6
+         , "0"      --  7
+         ]
+  , HU.testCase "HalfEven: negate r5" $ do
+       render Money.HalfEven (negate r5) @?=
+         [ "-0.50"   --  0
+         , "-0.50"   --  1
+         , "-0.50"   --  2
+         , "-0.50"   --  3
+         , "0"      --  4
+         , "0"      --  5
+         , "0"      --  6
+         , "0"      --  7
+         ]
+
   ]
   where
     r1 :: Rational = 1023004567895 % 1000
     r2 :: Rational = 123 % 100
     r3 :: Rational = 345 % 1000
     r4 :: Rational = 7 % 1000
+    r5 :: Rational = 1 % 2
 
     render :: Money.Approximation -> Rational -> [T.Text]
     render a r =
-      [ fromJust $ MoneyI.rationalToDecimal a False Nothing    '.' 2 r  --  0
-      , fromJust $ MoneyI.rationalToDecimal a False (Just ',') '.' 2 r  --  1
-      , fromJust $ MoneyI.rationalToDecimal a True  Nothing    '.' 2 r  --  2
-      , fromJust $ MoneyI.rationalToDecimal a True  (Just ',') '.' 2 r  --  3
-      , fromJust $ MoneyI.rationalToDecimal a False Nothing    '.' 0 r  --  4
-      , fromJust $ MoneyI.rationalToDecimal a False (Just ',') '.' 0 r  --  5
-      , fromJust $ MoneyI.rationalToDecimal a True  Nothing    '.' 0 r  --  6
-      , fromJust $ MoneyI.rationalToDecimal a True  (Just ',') '.' 0 r  --  7
+      [ MoneyI.rationalToDecimal ds_dd_2 a r        -- 0
+      , MoneyI.rationalToDecimal ds_tc_dd_2 a r     -- 1
+      , MoneyI.rationalToDecimal ds_p_dd_2 a r      -- 2
+      , MoneyI.rationalToDecimal ds_p_tc_dd_2 a r   -- 3
+      , MoneyI.rationalToDecimal ds_dd_0 a r        -- 4
+      , MoneyI.rationalToDecimal ds_tc_dd_0 a r     -- 5
+      , MoneyI.rationalToDecimal ds_p_dd_0 a r      -- 6
+      , MoneyI.rationalToDecimal ds_p_tc_dd_0 a r   -- 7
       ]
 
+testMkSeparators :: Tasty.TestTree
+testMkSeparators =
+  Tasty.testGroup "mkSeparators"
+  [ QC.testProperty "Unsupported separators" $
+      let mbadsep :: QC.Gen Char = QC.suchThat QC.arbitrary $ \c ->
+            Char.isDigit c || Char.isControl c
+          mgoodsep :: QC.Gen Char = QC.suchThat QC.arbitrary $ \c ->
+            not (Char.isDigit c || Char.isControl c)
+      in QC.forAll ((,,) <$> mbadsep <*> mbadsep <*> mgoodsep) $
+           \(s1 :: Char, s2 :: Char, s3 :: Char) ->
+               (Money.mkSeparators s1 Nothing   === Nothing) .&&.
+               (Money.mkSeparators s2 (Just s1) === Nothing) .&&.
+               (Money.mkSeparators s1 (Just s1) === Nothing) .&&.
+               (Money.mkSeparators s3 (Just s3) === Nothing)
+  ]
+
 testRationalFromDecimal :: Tasty.TestTree
 testRationalFromDecimal =
   Tasty.testGroup "rationalFromDecimal"
-  [ QC.testProperty "Unsupported separators" $
-      let mbadsep :: QC.Gen Char = QC.suchThat QC.arbitrary Char.isDigit
-          mgoodsep :: QC.Gen Char = QC.suchThat QC.arbitrary (not . Char.isDigit)
-      in QC.forAll ((,,) <$> mbadsep <*> mbadsep <*> mgoodsep) $
-           \(s1 :: Char, s2 :: Char, s3 :: Char) ->
-              let f = MoneyI.rationalFromDecimal
-              in (f Nothing s1 (error "untouched") === Nothing) .&&.
-                 (f (Just s1) s2 (error "untouched") === Nothing) .&&.
-                 (f (Just s1) s1 (error "untouched") === Nothing) .&&.
-                 (f (Just s3) s3 (error "untouched") === Nothing)
-
-  , Tasty.localOption (QC.QuickCheckTests 1000) $
+  [ Tasty.localOption (QC.QuickCheckTests 1000) $
     QC.testProperty "Lossy roundtrip" $
       -- We check that the roundtrip results in a close amount with a fractional
       -- difference of up to one.
-      let gen = (,) <$> genDecimalSeps <*> QC.arbitrary
-      in QC.forAll gen $ \( (yst :: Maybe Char, sd :: Char)
-                          , (r :: Rational, plus :: Bool, digs :: Word8,
-                             aprox :: Money.Approximation) ) ->
-           let Just dec = MoneyI.rationalToDecimal aprox plus yst sd digs r
-               Just r' = MoneyI.rationalFromDecimal yst sd dec
-           in 1 > abs (abs r - abs r')
+      QC.forAll QC.arbitrary $ \(ds0 :: Money.DecimalConf,
+                                 aprox :: Money.Approximation,
+                                 r :: Rational) ->
+        let Just sr = Money.scaleFromRational
+                    $ (+ 1) -- scales less than 1 can't reliably roundtrip.
+                    $ Money.scaleToRational
+                    $ Money.decimalConf_scale ds0
+            ds1 = ds0 { Money.decimalConf_scale = sr }
+            dec = MoneyI.rationalToDecimal ds1 aprox r
+            Just r' = MoneyI.rationalFromDecimal ds1 dec
+        in 1 > abs (abs r - abs r')
   ]
 
 testDense
@@ -489,42 +612,39 @@ testDense pc =
         T.unpack (Money.denseCurrency x) === symbolVal pc
 
   , QC.testProperty "denseToDecimal: Same as rationalToDecimal" $
-      let gen = (,) <$> genDecimalSeps <*> QC.arbitrary
-      in QC.forAll gen $ \( (yst :: Maybe Char, sd :: Char)
-                          , (dns :: Money.Dense currency, plus :: Bool,
-                             digs :: Word8, aprox :: Money.Approximation) ) ->
-            let ydnsd1 = Money.denseToDecimal aprox plus yst sd digs 1 dns
-                ydnsd100 = Money.denseToDecimal aprox plus yst sd digs 100 dns
-                yrd1 = MoneyI.rationalToDecimal aprox plus yst sd digs (toRational dns)
-                yrd100 = MoneyI.rationalToDecimal aprox plus yst sd digs (toRational dns * 100)
-            in (ydnsd1 === yrd1) .&&. (ydnsd100 === yrd100)
+      QC.forAll QC.arbitrary $ \(ds :: Money.DecimalConf,
+                                 dns :: Money.Dense currency,
+                                 aprox :: Money.Approximation) ->
+         let Just s100 = Money.scaleFromRational
+                    (Money.scaleToRational (Money.decimalConf_scale ds) * 100)
+             ds_s100 = ds { Money.decimalConf_scale = s100 }
+             ydnsd1 = Money.denseToDecimal ds aprox dns
+             ydnsd100 = Money.denseToDecimal ds_s100 aprox dns
+             yrd1 = MoneyI.rationalToDecimal ds aprox (toRational dns)
+             yrd100 = MoneyI.rationalToDecimal ds aprox (toRational dns * 100)
+         in (ydnsd1 === yrd1) .&&. (ydnsd100 === yrd100)
 
   , QC.testProperty "denseFromDecimal: Same as rationalFromDecimal" $
-      QC.forAll genDecimal $ \(dec :: T.Text, yts :: Maybe Char, ds :: Char) ->
-         let Just r = MoneyI.rationalFromDecimal yts ds dec
-             Just dns = Money.denseFromDecimal yts ds 1 dec
+      QC.forAll QC.arbitrary $ \(Decimal ds dec) ->
+         let Just r = MoneyI.rationalFromDecimal ds dec
+             Just dns = Money.denseFromDecimal ds dec
          in r === toRational (dns :: Money.Dense currency)
 
-  , QC.testProperty "denseFromDecimal: Same as rationalFromDecimal" $
-      QC.forAll genDecimal $ \(dec :: T.Text, yts :: Maybe Char, ds :: Char) ->
-         let Just r = MoneyI.rationalFromDecimal yts ds dec
-             Just dns = Money.denseFromDecimal yts ds 1 dec
-         in r === toRational (dns :: Money.Dense currency)
-
-
   , Tasty.localOption (QC.QuickCheckTests 1000) $
     QC.testProperty "denseToDecimal/denseFromDiscrete: Lossy roundtrip" $
       -- We check that the roundtrip results in a close amount with a fractional
       -- difference of up to one.
-      let gen = (,) <$> genDecimalSeps <*> QC.arbitrary
-      in QC.forAll gen $ \( (yst :: Maybe Char, sd :: Char)
-                          , (sc0 :: Rational, plus :: Bool,
-                             digs :: Word8, aprox :: Money.Approximation,
-                             dns :: Money.Dense currency) ) ->
-           let sc = abs sc0 + 1 -- smaller scales can't reliably be parsed back
-               Just dec = Money.denseToDecimal aprox plus yst sd digs sc dns
-               Just dns' = Money.denseFromDecimal yst sd sc dec
-           in Money.dense' 1 > abs (abs dns - abs dns')
+      QC.forAll QC.arbitrary $ \(ds0 :: Money.DecimalConf,
+                                 aprox :: Money.Approximation,
+                                 dns :: Money.Dense currency) ->
+        let Just sr = Money.scaleFromRational
+                    $ (+ 1) -- scales less than 1 can't reliably roundtrip.
+                    $ Money.scaleToRational
+                    $ Money.decimalConf_scale ds0
+            ds1 = ds0 { Money.decimalConf_scale = sr }
+            dec = Money.denseToDecimal ds1 aprox dns
+            Just dns' = Money.denseFromDecimal ds1 dec
+        in Money.dense' 1 > abs (abs dns - abs dns')
 
   , HU.testCase "AdditiveGroup: zeroV" $
       (AG.zeroV :: Money.Dense currency) @?= Money.dense' (0%1)
@@ -587,7 +707,7 @@ testExchange =
 
 testDiscrete
   :: forall (currency :: Symbol) (unit :: Symbol)
-  .  ( Money.GoodScale (Money.Scale currency unit)
+  .  ( Money.GoodScale (Money.UnitScale currency unit)
      , KnownSymbol currency
      , KnownSymbol unit )
   => Proxy currency
@@ -615,7 +735,7 @@ testDiscrete pc pu =
       QC.forAll QC.arbitrary $ \(dr :: Money.SomeDiscrete) ->
         ((T.unpack (Money.someDiscreteCurrency dr) /= symbolVal pc) &&
          (Money.someDiscreteScale dr /=
-             Money.scale (Proxy :: Proxy (Money.Scale currency unit)))
+             Money.scale (Proxy :: Proxy (Money.UnitScale currency unit)))
         ) ==> isNothing (Money.fromSomeDiscrete dr
                           :: Maybe (Money.Discrete currency unit))
   , QC.testProperty "withSomeDiscrete" $
@@ -633,16 +753,14 @@ testDiscrete pc pu =
   , QC.testProperty "discreteToDecimal/discreteFromDecimal: Same as denseToDecimal/denseFromDecimal" $
       -- We check that the roundtrip results in a close amount with a fractional
       -- difference of up to one.
-      let gen = (,) <$> genDecimalSeps <*> QC.arbitrary
-      in QC.forAll gen $ \( (yst :: Maybe Char, sd :: Char)
-                          , (sc0 :: Rational, plus :: Bool,
-                             digs :: Word8, aprox :: Money.Approximation,
-                             dis :: Money.Discrete currency unit) ) ->
-           let sc = abs sc0 + 1  -- scale can't be less than 1
-               dns = Money.denseFromDiscrete dis
-               ydec = Money.discreteToDecimal aprox plus yst sd digs sc dis
-               ydec' = Money.denseToDecimal aprox plus yst sd digs sc dns
-           in ydec === ydec'
+      QC.forAll QC.arbitrary $ \(ds :: Money.DecimalConf,
+                                 sc0 :: Rational,
+                                 aprox :: Money.Approximation,
+                                 dis :: Money.Discrete currency unit) ->
+        let sc = abs sc0 + 1  -- scales less than 1 can't reliably roundtrip.
+            dns = Money.denseFromDiscrete dis
+        in Money.discreteToDecimal ds aprox dis
+             === Money.denseToDecimal ds aprox dns
 
   , HU.testCase "AdditiveGroup: zeroV" $
       (AG.zeroV :: Money.Discrete currency unit) @?= Money.discrete 0
@@ -751,19 +869,17 @@ testExchangeRate ps pd =
              (show x, dr) === (show x', Money.toSomeExchangeRate x')
 
   , QC.testProperty "exchangeRateToDecimal: Same as rationalToDecimal" $
-      let gen = (,) <$> genDecimalSeps <*> QC.arbitrary
-      in QC.forAll gen $ \( (yst :: Maybe Char, sd :: Char)
-                          , (xr :: Money.ExchangeRate src dst, digs :: Word8,
-                             aprox :: Money.Approximation ) ) ->
-           let xrd = Money.exchangeRateToDecimal aprox yst sd digs xr
-               rd = MoneyI.rationalToDecimal aprox False yst sd digs
-                       (Money.exchangeRateToRational xr)
-           in xrd === rd
+      QC.forAll QC.arbitrary $ \(ds :: Money.DecimalConf,
+                                 xr :: Money.ExchangeRate src dst,
+                                 aprox :: Money.Approximation) ->
+        Money.exchangeRateToDecimal ds aprox xr
+          === MoneyI.rationalToDecimal ds aprox
+                (Money.exchangeRateToRational xr)
 
   , QC.testProperty "exchangeRateFromDecimal: Same as rationalFromDecimal" $
-      QC.forAll genDecimal $ \(dec :: T.Text, yts :: Maybe Char, ds :: Char) ->
-         let Just r = MoneyI.rationalFromDecimal yts ds dec
-             yxr = Money.exchangeRateFromDecimal yts ds dec
+      QC.forAll QC.arbitrary $ \(Decimal ds dec) ->
+         let Just r = MoneyI.rationalFromDecimal ds dec
+             yxr = Money.exchangeRateFromDecimal ds dec
                       :: Maybe (Money.ExchangeRate src dst)
          in (r > 0) ==> (Just r === fmap Money.exchangeRateToRational yxr)
 
@@ -793,57 +909,60 @@ testDiscreteFromDecimal :: Tasty.TestTree
 testDiscreteFromDecimal =
   Tasty.testGroup "discreteFromDecimal"
   [ HU.testCase "Too large" $ do
-      Money.discreteFromDecimal Nothing '.' 1 "0.053"
+      Money.discreteFromDecimal ds_dd_2 "0.053"
         @?= (Nothing :: Maybe (Money.Discrete "USD" "cent"))
-      Money.discreteFromDecimal (Just ',') '.' 1 "0.253"
+      Money.discreteFromDecimal ds_tc_dd_2 "0.253"
         @?= (Nothing :: Maybe (Money.Discrete "USD" "cent"))
 
   , HU.testCase "USD cent, small, zero" $ do
       let dis = 0 :: Money.Discrete "USD" "cent"
-          f = Money.discreteFromDecimal
-      f Nothing '.' 1 "0" @?= Just dis
-      f Nothing '.' 1 "+0" @?= Just dis
-      f Nothing '.' 1 "-0" @?= Just dis
-      f (Just ',') '.' 1 "0" @?= Just dis
-      f (Just ',') '.' 1 "+0" @?= Just dis
-      f (Just ',') '.' 1 "-0" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2       "0" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "+0" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "-0" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2     "0" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "+0" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "-0" @?= Just dis
 
   , HU.testCase "USD cent, small, positive" $ do
       let dis = 25 :: Money.Discrete "USD" "cent"
-          f = Money.discreteFromDecimal
-      f Nothing '.' 1 "0.25" @?= Just dis
-      f Nothing '.' 1 "+0.25" @?= Just dis
-      f (Just ',') '.' 1 "0.25" @?= Just dis
-      f (Just ',') '.' 1 "+0.25" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "0.25"  @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "+0.25" @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "0.25"  @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "+0.25" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "0.25"  @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "+0.25" @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "0.25"  @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "+0.25" @?= Just dis
 
   , HU.testCase "USD cent, small, negative" $ do
       let dis = -25 :: Money.Discrete "USD" "cent"
-          f = Money.discreteFromDecimal
-      f Nothing '.' 1 "-0.25" @?= Just dis
-      f Nothing '.' 1 "-0.25" @?= Just dis
-      f (Just ',') '.' 1 "-0.25" @?= Just dis
-      f (Just ',') '.' 1 "-0.25" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "-0.25" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "-0.25" @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "-0.25" @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "-0.25" @?= Just dis
 
   , HU.testCase "USD cent, big, positive" $ do
       let dis = 102300456789 :: Money.Discrete "USD" "cent"
-          f = Money.discreteFromDecimal
-      f Nothing '.' 1 "1023004567.89" @?= Just dis
-      f Nothing '.' 1 "+1023004567.89" @?= Just dis
-      f (Just ',') '.' 1 "1,023,004,567.89" @?= Just dis
-      f (Just ',') '.' 1 "+1,023,004,567.89" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "+1023004567.89"    @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "1023004567.89"     @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "+1023004567.89"    @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "1023004567.89"     @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "+1,023,004,567.89" @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "1,023,004,567.89"  @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "+1,023,004,567.89" @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "1,023,004,567.89"  @?= Just dis
 
   , HU.testCase "USD cent, big, negative" $ do
       let dis = -102300456789 :: Money.Discrete "USD" "cent"
-          f = Money.discreteFromDecimal
-      f Nothing '.' 1 "-1023004567.89" @?= Just dis
-      f Nothing '.' 1 "-1023004567.89" @?= Just dis
-      f (Just ',') '.' 1 "-1,023,004,567.89" @?= Just dis
-      f (Just ',') '.' 1 "-1,023,004,567.89" @?= Just dis
+      Money.discreteFromDecimal ds_dd_2      "-1023004567.89" @?= Just dis
+      Money.discreteFromDecimal ds_p_dd_2    "-1023004567.89" @?= Just dis
+      Money.discreteFromDecimal ds_tc_dd_2   "-1,023,004,567.89" @?= Just dis
+      Money.discreteFromDecimal ds_p_tc_dd_2 "-1,023,004,567.89" @?= Just dis
   ]
 
 testRounding
   :: forall (currency :: Symbol) (unit :: Symbol)
-  .  (Money.GoodScale (Money.Scale currency unit), KnownSymbol currency)
+  .  (Money.GoodScale (Money.UnitScale currency unit), KnownSymbol currency)
   => Proxy currency
   -> Proxy unit
   -> Tasty.TestTree
@@ -853,24 +972,98 @@ testRounding _ _ =
     , QC.testProperty "ceiling"  $ QC.forAll QC.arbitrary (g (Money.discreteFromDense Money.Ceiling))
     , QC.testProperty "round"    $ QC.forAll QC.arbitrary (g (Money.discreteFromDense Money.Round))
     , QC.testProperty "truncate" $ QC.forAll QC.arbitrary (g (Money.discreteFromDense Money.Truncate))
+    , QC.testProperty "half-even" $ QC.forAll QC.arbitrary (g (Money.discreteFromDense Money.HalfEven))
     , QC.testProperty "floor no reminder"    $ QC.forAll QC.arbitrary (h (Money.discreteFromDense Money.Floor))
     , QC.testProperty "ceiling no reminder"  $ QC.forAll QC.arbitrary (h (Money.discreteFromDense Money.Ceiling))
     , QC.testProperty "round no reminder"    $ QC.forAll QC.arbitrary (h (Money.discreteFromDense Money.Round))
     , QC.testProperty "truncate no reminder" $ QC.forAll QC.arbitrary (h (Money.discreteFromDense Money.Truncate))
+    , QC.testProperty "half-even no reminder " $ QC.forAll QC.arbitrary (h (Money.discreteFromDense Money.HalfEven))
     ]
   where
-    g :: (Money.Dense currency -> (Money.Discrete' currency (Money.Scale currency unit), Money.Dense currency))
+    g :: (Money.Dense currency -> (Money.Discrete' currency (Money.UnitScale currency unit), Money.Dense currency))
       -> Money.Dense currency
       -> QC.Property
     g f = \x -> x === case f x of (y, z) -> Money.denseFromDiscrete y + z
 
-    h :: (Money.Dense currency -> (Money.Discrete' currency (Money.Scale currency unit), Money.Dense currency))
+    h :: (Money.Dense currency -> (Money.Discrete' currency (Money.UnitScale currency unit), Money.Dense currency))
       -> Money.Discrete currency unit
       -> QC.Property
     h f = \x -> (Money.denseFromDiscrete x) === case f (Money.denseFromDiscrete x) of
       (y, 0) -> Money.denseFromDiscrete y
       (_, _) -> error "testRounding.h: unexpected"
 
+testHalfEvenRounding :: Tasty.TestTree
+testHalfEvenRounding =
+    Tasty.testGroup "HalfEven rounding"
+    [ HU.testCase "-2.8" $ -3 @=? f (-3152519739159347 % 1125899906842624)
+
+    , HU.testCase "-2.5" $ -2 @=? f (-5 % 2)
+    , HU.testCase "-2.2" $ -2 @=? f (-2476979795053773 % 1125899906842624)
+    , HU.testCase "-2"   $ -2 @=? f (-2 % 1)
+    , HU.testCase "-1.8" $ -2 @=? f (-8106479329266893 % 4503599627370496)
+    , HU.testCase "-1.5" $ -2 @=? f (-3 % 2)
+
+    , HU.testCase "-1.2" $ -1 @=? f (-5404319552844595 % 4503599627370496)
+    , HU.testCase "-1"   $ -1 @=? f (-1 % 1)
+    , HU.testCase "-0.8" $ -1 @=? f (-3602879701896397 % 4503599627370496)
+
+    , HU.testCase "-0.5" $  0 @=? f (-1 % 2)
+    , HU.testCase "-0.2" $  0 @=? f (-3602879701896397 % 18014398509481984)
+    , HU.testCase  "0"   $  0 @=? f (0 % 1)
+    , HU.testCase  "0.2" $  0 @=? f (3602879701896397 % 18014398509481984)
+    , HU.testCase  "0.5" $  0 @=? f (1 % 2)
+
+    , HU.testCase  "0.8" $  1 @=? f (3602879701896397 % 4503599627370496)
+    , HU.testCase  "1"   $  1 @=? f (1 % 1)
+    , HU.testCase  "1.2" $  1 @=? f (5404319552844595 % 4503599627370496)
+
+    , HU.testCase  "1.5" $  2 @=? f (3 % 2)
+    , HU.testCase  "1.8" $  2 @=? f (8106479329266893 % 4503599627370496)
+    , HU.testCase  "2"   $  2 @=? f (2 % 1)
+    , HU.testCase  "2.2" $  2 @=? f (2476979795053773 % 1125899906842624)
+    , HU.testCase  "2.5" $  2 @=? f (5 % 2)
+
+    , HU.testCase  "2.8" $  3 @=? f (3152519739159347 % 1125899906842624)
+    ]
+  where
+    f :: Rational -> Integer
+    f = MoneyI.approximate MoneyI.HalfEven
+
+
+
+-- | Decimal dot, 2 decimals
+ds_dd_2 :: Money.DecimalConf
+ds_dd_2 = Money.defaultDecimalConf
+
+-- | Leading plus, decimal dot, 2 decimals
+ds_p_dd_2 :: Money.DecimalConf
+ds_p_dd_2 = ds_dd_2 { Money.decimalConf_leadingPlus = True }
+
+-- | Thousands comma, decimal dot, 2 decimals
+ds_tc_dd_2 :: Money.DecimalConf
+ds_tc_dd_2 = ds_dd_2 { Money.decimalConf_separators = Money.separatorsDotComma }
+
+-- | Leading plus, thousands comma, decimal dot, 2 decimals
+ds_p_tc_dd_2 :: Money.DecimalConf
+ds_p_tc_dd_2 = ds_tc_dd_2 { Money.decimalConf_leadingPlus = True }
+
+-- | Decimal dot, 0 decimals
+ds_dd_0 :: Money.DecimalConf
+ds_dd_0 = ds_dd_2 { Money.decimalConf_digits = 0 }
+
+-- | Leading plus, decimal dot, 0 decimals
+ds_p_dd_0 :: Money.DecimalConf
+ds_p_dd_0 = ds_p_dd_2 { Money.decimalConf_digits = 0 }
+
+-- | Thousands comma, decimal dot, 0 decimals
+ds_tc_dd_0 :: Money.DecimalConf
+ds_tc_dd_0 = ds_tc_dd_2 { Money.decimalConf_digits = 0 }
+
+-- | Leading plus, thousands comma, decimal dot, 0 decimals
+ds_p_tc_dd_0 :: Money.DecimalConf
+ds_p_tc_dd_0 = ds_p_tc_dd_2 { Money.decimalConf_digits = 0 }
+
+
 --------------------------------------------------------------------------------
 -- Raw parsing "golden tests"
 
@@ -914,6 +1107,16 @@ rawDis0_binary = "\NUL\NUL\NUL\NUL\NUL\NUL\NUL\ETXUSD\NUL\NUL\NUL\NULd\NUL\NUL\N
 rawXr0_binary :: BL.ByteString
 rawXr0_binary = "\NUL\NUL\NUL\NUL\NUL\NUL\NUL\ETXUSD\NUL\NUL\NUL\NUL\NUL\NUL\NUL\ETXBTC\NUL\NUL\NUL\NUL\ETX\NUL\NUL\NUL\NUL\STX"
 
+--------------------------------------------------------------------------------
+
+-- Just checking that this compiles.
+_testCurrencyScale
+  :: Dict (Money.CurrencyScale "USD" ~ Money.UnitScale "USD" "cent",
+           Money.CurrencyScale "BTC" ~ Money.UnitScale "BTC" "satoshi",
+           Money.CurrencyScale "VUV" ~ Money.UnitScale "VUV" "vatu",
+           Money.CurrencyScale "XAU" ~ Money.ErrScaleNonCanonical "XAU")
+_testCurrencyScale = Dict
+
 --------------------------------------------------------------------------------
 -- Misc
 
